import inspect
import json
import re
from typing import TYPE_CHECKING, Iterator

import kubernetes
from kubernetes.client.rest import ApiException

from pykorm.pykorm import Pykorm

if TYPE_CHECKING:
    from .models import PykormModel, NamespacedModel, ClusterModel


class Filter:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


class Node(object):
    def clone(self):
        obj = self.__class__.__new__(self.__class__)
        obj.__dict__ = self.__dict__.copy()
        # We also copy the following attrs in order to not_impact other instances
        obj._next_filter = obj._next_filter.copy()
        obj._label_filters = obj._label_filters.copy()
        obj._field_filters = obj._field_filters.copy()

        return obj

    @staticmethod
    def copy(method):
        def inner(self, *args, **kwargs):
            clone = self.clone()
            method(clone, *args, **kwargs)
            return clone

        return inner


class BaseQuery(Node):
    baseobject: 'PykormModel'

    def __init__(self, baseobject: 'PykormModel', api):
        self.baseobject = baseobject
        self._next_filter = []
        self._label_filters = {}
        self._field_filters = {}
        self._api = api
        self._current_cluster = 'default'
        self._overwrite_api_client = None
        self._overwrite_api_client_mapping = {}

    @property
    def api_client_mapping(self):
        if self._overwrite_api_client_mapping:
            return self._overwrite_api_client_mapping
        return Pykorm.get_api_client_mapping()

    def _reset_api_client(self):
        self._api.api_client = self.api_client

    @Node.copy
    def overwrite_clusters(self, clusters_config):
        self._overwrite_api_client_mapping = Pykorm.get_api_client_mapping(clusters_config)
        self._reset_api_client()

    @property
    def api_client(self):
        if self._overwrite_api_client:
            return self._overwrite_api_client
        return self.api_client_mapping[self._current_cluster]

    @property
    def api(self):
        if inspect.isclass(self._api):
            self._api = self._api(self.api_client)
        return self._api

    @property
    def is_crd(self):
        return isinstance(self.api, kubernetes.client.CustomObjectsApi)

    def __call__(self, *args, **kwargs):
        if 'overwrite_api_client' in kwargs:
            self._overwrite_api_client = kwargs['overwrite_api_client']
            self._reset_api_client()
        return self.__class__(self.baseobject, self.api)

    @property
    def api_resource_name(self):
        if self.is_crd:
            return 'custom_object'
        found = re.findall('[A-Z][^A-Z]*', self.baseobject._pykorm_kind)
        return '_'.join([_.lower() for _ in found])

    def process_method_kwargs(self, obj: 'PykormModel', with_labels=False, **kwargs):
        from .models import NamespacedModel
        ret = kwargs

        if isinstance(obj, NamespacedModel):
            ret['namespace'] = obj.namespace

        if self.is_crd:
            ret.update({
                'group': obj._pykorm_group,
                'version': obj._pykorm_version,
                'plural': obj._pykorm_plural,
            })
        if with_labels:
            labels = {}
            labels.update(self._label_filters)  # from query
            ret['label_selector'] = obj.process_labels(labels)

            fields = self._field_filters
            ret['field_selector'] = obj.process_labels(fields)

        ret['_preload_content'] = False
        ret['_request_timeout'] = (2, 5)
        return ret

    @Node.copy
    def filter_by(self, **kwargs: str):
        self._next_filter.append(Filter(**kwargs))

    @Node.copy
    def filter_by_labels(self, **kwargs):
        self._label_filters.update(kwargs)

    @Node.copy
    def filter_by_fields(self, **kwargs):
        self._field_filters.update(kwargs)

    @Node.copy
    def using(self, cluster='default'):
        self._current_cluster = cluster
        self._reset_api_client()

    def _query(self, data=None, **kwargs):
        ret = []
        if data is None:
            data = self._iter(**kwargs)
        for el in data:
            if el._matches_attributes(kwargs):
                ret.append(el)
        return ret

    def all(self) -> ['PykormModel']:
        last_ret = None
        if not self._next_filter:
            return self._query()
        for _filter in self._next_filter:
            last_ret = self._query(data=last_ret, **_filter.kwargs)
        return last_ret

    def first(self):
        _all = self.all()
        if _all:
            return _all[0]

    def get(self, name):
        all_kwargs = {}
        [all_kwargs.update(f.kwargs) for f in self._next_filter]
        namespace = all_kwargs.get('namespace')
        if not namespace:
            raise Exception('Must specify namespace')
        return self._get(namespace, name)

    def _save(self, obj: 'PykormModel'):
        k8s_dict = obj._k8s_dict

        kwargs = self.process_method_kwargs(obj, body=k8s_dict)

        if obj._k8s_uid is None:
            result = self.create_method(**kwargs)
        else:
            kwargs['name'] = obj.name
            result = self.patch_method(**kwargs)

        obj._set_attributes_with_dict(self.process_http_response(result))
        obj._queryset = self

    def _apply(self, obj: 'PykormModel'):
        k8s_dict = obj._k8s_dict

        try:
            result = self.create_method(**self.process_method_kwargs(obj, body=k8s_dict))
        except ApiException as e:
            if e.status == 404:
                result = self.patch_method(**self.process_method_kwargs(obj, body=k8s_dict))

        obj._set_attributes_with_dict(self.process_http_response(result))

    @staticmethod
    def process_http_response(http_response):
        return json.loads(http_response.read())


class NamespacedObjectQuery(BaseQuery):
    def _iter(self, **kwargs) -> Iterator['NamespacedModel']:
        base_cls = self.baseobject
        ns_name = kwargs.get('namespace', None)
        namespaces = {kwargs.get('namespace', 'default')}

        if not ns_name:
            corev1 = kubernetes.client.CoreV1Api(self.api_client)
            for namespace in corev1.list_namespace().items:
                ns_name = namespace.metadata.name
                namespaces.add(ns_name)
        for ns_name in namespaces:
            objs = self.process_http_response(
                self.list_method(**self.process_method_kwargs(base_cls, with_labels=True, namespace=ns_name)))
            for obj in objs['items']:
                yield self.baseobject._instantiate_with_dict(obj, queryset=self)

    @property
    def get_method(self):
        if self.is_crd:
            return getattr(self.api, f'get_namespaced_{self.api_resource_name}')
        return getattr(self.api, f'read_namespaced_{self.api_resource_name}')

    @property
    def list_method(self):
        return getattr(self.api, f'list_namespaced_{self.api_resource_name}')

    @property
    def patch_method(self):
        return getattr(self.api, f'patch_namespaced_{self.api_resource_name}')

    @property
    def create_method(self):
        return getattr(self.api, f'create_namespaced_{self.api_resource_name}')

    @property
    def delete_method(self):
        return getattr(self.api, f'delete_namespaced_{self.api_resource_name}')

    def _get(self, namespace, name):
        try:
            result = self.get_method(
                **self.process_method_kwargs(self.baseobject, namespace=namespace, name=name)
            )
        except ApiException as e:
            if e.status == 404:
                return
            else:
                raise Exception
        if result:
            obj = self.process_http_response(result)
            return self.baseobject._instantiate_with_dict(obj, queryset=self)

    def _delete(self, obj: 'NamespacedModel'):
        self.delete_method(**self.process_method_kwargs(obj, namespace=obj.namespace, name=obj.name))


class ClusterObjectQuery(BaseQuery):
    @property
    def get_method(self):
        if self.is_crd:
            return self.api.get_cluster_custom_object
        return getattr(self.api, f'get_{self.api_resource_name}')

    @property
    def list_method(self):
        if self.is_crd:
            return self.api.list_cluster_custom_object
        return getattr(self.api, f'list_{self.api_resource_name}')

    @property
    def patch_method(self):
        if self.is_crd:
            return self.api.patch_cluster_custom_object
        return getattr(self.api, f'patch_{self.api_resource_name}')

    @property
    def create_method(self):
        if self.is_crd:
            return self.api.create_cluster_custom_object
        return getattr(self.api, f'create_{self.api_resource_name}')

    @property
    def delete_method(self):
        if self.is_crd:
            return self.api.delete_cluster_custom_object
        return getattr(self.api, f'delete_{self.api_resource_name}')

    def _iter(self, **kwargs) -> Iterator['ClusterModel']:
        base_cls = self.baseobject
        objs = self.process_http_response(self.list_method(**self.process_method_kwargs(base_cls, with_labels=True)))
        for obj in objs['items']:
            yield self.baseobject._instantiate_with_dict(obj, queryset=self)

    def _delete(self, obj: 'ClusterModel'):
        self.delete_method(**self.process_method_kwargs(obj, name=obj.name))

    def get(self, name):
        return self._get(name)

    def _get(self, name):
        try:
            result = self.get_method(**self.process_method_kwargs(self.baseobject, name=name))
        except ApiException as e:
            if e.status == 404:
                return
        if result:
            obj = self.process_http_response(result)
            return self.baseobject._instantiate_with_dict(obj, queryset=self)



from typing import Dict

import dpath.util

from . import fields
from . import query as pykorm_query
from .meta import ModelMixin

IGNORE_HIDDEN_ATTRIBUTE = True


class PykormModel(ModelMixin):
    name: str = fields.Metadata('name', readonly=True)
    created_at: str = fields.Metadata('creationTimestamp', readonly=True)
    _k8s_uid: str = fields.Metadata('uid', readonly=True)

    _pykorm_group: str
    _pykorm_version: str
    _pykorm_plural: str
    _pykorm_kind: str
    _property_fields: dict = {}

    _queryset: pykorm_query.BaseQuery

    @classmethod
    def process_labels(cls, d: Dict[str, str]) -> str:
        ''' Converts a labels_dict to a k=v,k=v string '''
        return ','.join([f'{k}={v}' for k, v in d.items()])

    def to_dict(self):
        data = {}
        for attr_name, _ in self._get_attributes().items():
            v = getattr(self, attr_name)
            if v is None:
                continue
            data[attr_name] = v
        for attr_name, _attr_method in self._property_fields.items():
            data[attr_name] = getattr(self, attr_name)

        return {k: v for k, v in data.items() if not (k.startswith('_') and IGNORE_HIDDEN_ATTRIBUTE)}

    def _matches_attributes(self, filters_dict: Dict[str, str]) -> bool:
        for attribute_name, attribute_value in filters_dict.items():
            if getattr(self, attribute_name) != attribute_value:
                return False
        return True

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self._k8s_dict == other._k8s_dict


    def __hash__(self):
        return hash(self._k8s_uid)



    @classmethod
    def _instantiate_with_dict(cls, k8s_dict, queryset) -> 'PykormModel':
        ''' Creates the model with data from the k8s data structure '''
        obj = cls.__new__(cls)
        obj._set_attributes_with_dict(k8s_dict)
        obj._queryset = queryset
        return obj


    def _set_attributes_with_dict(self, k8s_dict: Dict):
        self.__k8s_data = k8s_dict
        self._k8s_dict_cached = {}

        for (attr_name, attr) in self._get_attributes().items():
            value = attr.get_data(k8s_dict)
            self.__dict__[attr_name] = value


    @property
    def _k8s_dict(self):
        '''
        Returns the model as a kubernetes dict/yaml structure. This function
        does NOT return readonly fields
        '''
        if self._pykorm_group:
            api_version = f'{self._pykorm_group}/{self._pykorm_version}'
        else:
            api_version = self._pykorm_version

        d = {
            "apiVersion": api_version,
            "kind": self._pykorm_kind or self.__class__.__name__,
            "metadata": {
                "name": self.name,
            },
        }

        for (attr_name, attr_type) in self._get_attributes().items():
            attr_value = getattr(self, attr_name)

            if attr_type.required and not attr_value:
                raise Exception(f"Field {attr_name} must be set")

            if attr_type.readonly:  # Skip readonly field
                continue

            attr_dict_path = attr_type.to_dict(attr_value)
            d = dpath.util.merge(attr_dict_path, d)

        return d

    def delete(self):
        return self._queryset._delete(self)

    def save(self):
        return self._queryset._save(self)

    def apply(self):
        return self._queryset._apply(self)


class NamespacedModel(PykormModel):
    namespace = fields.Metadata('namespace')


class ClusterModel(PykormModel):
    pass




from typing import Dict

import dpath.util


class DataField:
    _root_dict_key = []
    path: str
    default: object
    readonly: bool = False
    required: bool = False
    allow_empty: bool = False

    def __init__(self, path: str, default=None, required: bool = False,
                 readonly: bool = False, allow_empty: bool = False):
        if type(path) == str:
            self.path = [path]
        elif type(path) == list:
            self.path = path
        else:
            raise Exception('param path must be a string or list')

        self.default = default
        self.required = required
        self.readonly = readonly
        self.allow_empty = allow_empty

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f'<pykorm/{cls_name}: {self.path}>'

    @property
    def fullpath(self):
        return self._root_dict_key + self.path

    @staticmethod
    def preprocessor(x):
        return x

    @staticmethod
    def postprocessor(x):
        return x

    def to_dict(self, value: str) -> Dict:
        result = {}
        if value is None and self.default is not None:
            value = self.default
        if value is None and (not self.allow_empty):
            return {}
        if value is not None:
            value = self.preprocessor(value)

        return dpath.util.new(result, self.fullpath, value)

    def get_data(self, k8s_dict: Dict):
        data = dpath.util.get(k8s_dict, self.fullpath, default=self.default)
        return self.postprocessor(data) if data is not None else data


class Spec(DataField):
    _root_dict_key = ['spec']


class Metadata(DataField):
    _root_dict_key = ['metadata']


class MetadataAnnotation(DataField):
    _root_dict_key = ['metadata', 'annotations']


class Status(DataField):
    _root_dict_key = 'status'



import kubernetes

from . import models
from . import query as pykorm_query


def _process_cls(cls, query_class, group: str, version: str, plural: str, api,
                 kind: str = None):
    cls._pykorm_group = group
    cls._pykorm_version = version
    cls._pykorm_plural = plural
    cls._pykorm_kind = kind
    cls.query = query_class(baseobject=cls, api=api)()
    return cls


def get_query_cls(cls):
    if issubclass(cls, models.ClusterModel):
        return pykorm_query.ClusterObjectQuery
    elif issubclass(cls, models.NamespacedModel):
        return pykorm_query.NamespacedObjectQuery
    else:
        raise Exception(f"Class {cls} doesn't seem to inherit from either ClusterModel nor NamespacedModel")


def k8s_custom_object(group: str, version: str, plural: str, kind: str = None):
    def wrap(cls):
        query_class = get_query_cls(cls)
        return _process_cls(cls, query_class, group, version, plural, api=kubernetes.client.CustomObjectsApi, kind=kind)

    return wrap


def k8s_core(kind: str):
    def wrap(cls):
        query_class = get_query_cls(cls)
        return _process_cls(cls, query_class, group=None, version='v1', plural=None, api=kubernetes.client.CoreV1Api,
                            kind=kind)

    return wrap


def k8s_appsv1(kind: str):
    def wrap(cls):
        query_class = get_query_cls(cls)
        return _process_cls(cls, query_class, group='apps', version='v1', plural=None,
                            api=kubernetes.client.AppsV1Api,
                            kind=kind)

    return wrap


class Pykorm:
    clusters_config = {}  # define api configuration

    @classmethod
    def get_api_client_mapping(cls, clusters_config=None):
        if clusters_config is None:  # allow overwrite global clusters config
            clusters_config = cls.clusters_config

        api_client_mapping = {}  # store api client mapping


        # Load the `default` cluster from kubeconfig/incluster config
        try:
            # To be changed once https://github.com/kubernetes-client/python/issues/1005 is fixed
            kubernetes.config.load_kube_config()
        except kubernetes.config.config_exception.ConfigException:
            kubernetes.config.load_incluster_config()

        api_client = kubernetes.client.ApiClient()
        api_client_mapping['default'] = api_client

        # Load other cluster if specified in configuration
        if clusters_config:
            for cluster_name, api_config in clusters_config.items():
                configuration = kubernetes.client.Configuration()
                for k, v in api_config.items():
                    setattr(configuration, k, v)

                api_client_mapping[cluster_name] = kubernetes.client.ApiClient(configuration)
        return api_client_mapping

    def save(self, obj: 'models.PykormModel', cluster: str = 'default'):
        obj.query().using(cluster)._save(obj)

    def delete(self, obj: 'models.PykormModel', cluster: str = 'default'):
        obj.query().using(cluster)._delete(obj)

    def apply(self, obj: 'models.PykormModel', cluster: str = 'default'):
        raise Exception('Do not use')
        obj.query().using(cluster)._apply(obj)



import inspect
from typing import Dict

from . import fields
from .exceptions import ReadOnlyException


class ModelMeta(type):
    def __new__(cls, name, bases, attrs):
        fields = {}
        for k, v in attrs.items():
            if isinstance(v, property) and not k.startswith('_'):
                fields[k] = v
        for base in bases:
            if hasattr(base, "_property_fields"):
                fields.update(base._property_fields)
        attrs["_property_fields"] = fields

        return type.__new__(cls, name, bases, attrs)


class ModelMixin(metaclass=ModelMeta):
    _attributes_cached = {}

    def __init__(self, *args, **kwargs):
        attrs = self._get_attributes()
        for k, v in kwargs.items():
            if k not in attrs and not isinstance(kwargs[k], fields.DataField) and k not in self._property_fields:
                raise Exception(f'{k} not in attrs')
            setattr(self, k, v)

    @classmethod
    def _get_attributes(cls) -> Dict[str, fields.DataField]:
        if cls._attributes_cached:
            return cls._attributes_cached
        attributes = inspect.getmembers(cls, lambda a: not (inspect.isroutine(a)))
        obj_attrs = [a for a in attributes if not (a[0].startswith('__') and a[0].endswith('__'))]

        retl = {}
        for obj in obj_attrs:
            (_attr_name, attr) = obj
            if isinstance(attr, fields.DataField):
                retl[_attr_name] = attr
        cls._attributes_cached = retl
        return retl

    def __setattr__(self, item: str, value):
        attrs = self._get_attributes()
        if item in attrs:
            attr = attrs[item]
            current_val = getattr(self, item)

            if attr.readonly and current_val is not None:
                # We allow to set the attribute if it was not set before
                if current_val != value:
                    raise ReadOnlyException(f'{item} attribute is read_only !')

        return super().__setattr__(item, value)

    def __repr__(self):
        data = self.to_dict()
        repr_str = [f'{k}={v}' for k, v in data.items()]
        return f'<{self.__class__.__name__} {" ".join(repr_str)}>'

    def __getattribute__(self, item: str):
        attr = object.__getattribute__(self, item)

        if isinstance(attr, fields.DataField):
            return attr.default
        else:
            return attr



import dpath

from .fields import Spec, Metadata, MetadataAnnotation  # noqa
from .models import NamespacedModel, ClusterModel  # noqa
from .pykorm import k8s_custom_object, k8s_core, Pykorm  # noqa

dpath.options.ALLOW_EMPTY_STRING_KEYS = True
